{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
{-# LANGUAGE FlexibleInstances, MultiParamTypeClasses, OverlappingInstances #-}

module F4L.Actions.Common where


import F4L
import F4L.F4LMonad
import F4L.FSMonad hiding ( outputLineMsg, cleanLine )
import qualified F4L.FSMonad
import F4L.Knowledge
import F4L.Op
import F4L.Utils hiding( fileSize )

import Control.Lens
import Control.Monad
import Control.Monad.Error( ErrorT(..) )
import Control.Monad.Reader
import Data.Function( on )
import qualified Data.Map.Strict as SM
import Data.Monoid( mappend )
import System.FilePath( combine, splitDirectories )


class VerboseMonad m where
  say :: VerbosityLevel -> String -> m ()
  outputLineMsg :: String -> m ()
  cleanLine :: m ()


instance FSMonad m => VerboseMonad ( GenF4LMonad k s m ) where
  say l msg = do
    v <- use $ st_verbosity . verb_level
    t <- use $ st_verbosity . verb_tracing
    when ( l <= v ) $ when t cleanLine >> outputMsg msg

  outputLineMsg msg =
    whenM ( use $ st_verbosity . verb_tracing ) $ F4L.FSMonad.outputLineMsg msg

  cleanLine =
    whenM ( use $ st_verbosity . verb_tracing ) F4L.FSMonad.cleanLine


-- instance ( Monad m, VerboseMonad m ) => VerboseMonad ( ErrorT Error m ) where
instance ( MonadTrans t, Monad m, VerboseMonad m ) => VerboseMonad ( t m ) where
  say l msg = lift $ say l msg
  outputLineMsg = lift . outputLineMsg
  cleanLine = lift cleanLine


sidPathCmp :: ( StorageId, FilePath ) -> ( StorageId, FilePath ) -> Ordering
sidPathCmp = ( compare `on` fst ) `mappend` ( compare `on` ( splitDirectories . snd ) )


-- | Force knowledge type within a GenF4LMonad based on first argument type
--   (value is ignored)
withKnowledge :: Monad m => k -> GenF4LMonad k s m ()
withKnowledge = const $ return ()


onErrorDie :: Monad m => ErrorT Error m a -> m a
onErrorDie a = do
  res <- runErrorT a
  case res of
    Left err -> error err
    Right x -> return x


onErrorUseDefault :: FSMonad m => a -> ErrorT Error m a -> m a
onErrorUseDefault def a = liftM ( either ( const def ) id ) $ runErrorT a


onErrorWarn :: ( VerboseMonad m, FSMonad m ) => ErrorT Error m () -> m ()
onErrorWarn = onErrorWarnWithDefault ()


onErrorWarnWithDefault :: ( VerboseMonad m, FSMonad m ) => a -> ErrorT Error m a -> m a
onErrorWarnWithDefault def a = do
  res <- runErrorT a
  case res of
    Left err -> say Quiet ( "Warning: " ++ err ) >> return def
    Right x -> return x


-- | Rich data structure for path within a storage
data RichPath = RichPath {
    rp_sid :: !StorageId -- ^ Storage ID
  , rp_sPath :: !FilePath -- ^ Storage path
  , rp_relPath :: !FilePath -- ^ Path wrt storage
  , rp_canPath :: !FilePath -- ^ Canonical path
  } deriving Show


{---------------------------------------- Actions ----------------------------------------}


interactivePrompt def msg = do
  i <- use st_interactive
  case i of
    Ask -> prompt msg
    UseDefault -> return def
    Use answer -> st_interactive .= UseDefault >> return answer


-- | Add component to RichPath
subRichPath :: FilePath -> RichPath -> RichPath
subRichPath fp rp = rp { rp_relPath = combine ( rp_relPath rp ) fp
                       , rp_canPath = combine ( rp_canPath rp ) fp
                       }

f4lState :: Knowledge k 
            => MetaDB -> Verbosity -> Interactivity -> Bool -> Bool -> s -> F4LState k s
f4lState a b c d e s = F4LState a b c d e initialOpEnv s SM.empty 0 noKnowledge

everyN :: Monad m => Int -> F4LMonad k m () -> F4LMonad k m ()
everyN n act = do
  i <- use st_debugCounter
  if i >= n - 1 then
    st_debugCounter .= 0 >> act
    else st_debugCounter += 1
