{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
module F4L.Actions.Ops (
  addOp
  , attrOp
  , extraSession
  , fromDB
  , getDBFile
  , importExportSessions
  , initDB
  , inSession
  , labelEnv
  , labelStorage
  , loadCached
  , loadOps
  , loadOpsDo
  , mergeDBs
  , readVOps
  , saveCache
  , sessions
  , setFid
  , setPath
  ) where

import F4L
import F4L.Actions.Common
import F4L.BufferMonad
import F4L.F4LMonad
import F4L.FSMonad hiding ( outputLineMsg, cleanLine )
import F4L.Knowledge
import F4L.Op
import F4L.Utils hiding( fileSize )

import qualified Codec.Compression.GZip as GZ
import Control.DeepSeq
import Control.Lens
import Control.Monad
import Control.Monad.Error( ErrorT(..) )
import Control.Monad.Reader
import Control.Monad.State.Strict
import Data.Array.IArray( elems )
import Data.Binary( encode )
import qualified Data.Binary as B
import Data.Binary.Get
import Data.Binary.Put( runPut )
import qualified Data.ByteString.Lazy as BL
import qualified Data.ByteString.Short as SBS
import Data.Function( on )
import Data.List( sortBy, unfoldr )
import System.FilePath( combine, takeDirectory )


execOp :: ( Monad m, Knowledge k ) => Op -> GenF4LMonad k s m [Fact]
execOp op = do
  openv <- use st_opEnv
  let ( facts, newOpenv ) = runState ( applyOp op ) openv
  st_opEnv .= newOpenv
  forM_ facts $ \ fact -> st_knowledge %= applyFact fact
  return facts


addOp :: ( BufferMonad m, Knowledge k ) => Op -> F4LMonad k m ()
addOp op = addOp' op >> return ()


addOp' :: ( BufferMonad m, Knowledge k ) => Op -> F4LMonad k m [Fact]
addOp' op = do
  addToBuffer $ encode op
  execOp op


setPath :: ( BufferMonad m, Knowledge k ) => FilePath -> F4LMonad k m ()
setPath path = do
  curpath <- use $ st_opEnv . openv_path
  unless ( curpath == path ) $ do
    mfid <- use $ st_opEnv . openv_fid
    case mfid of
         Just fid | path == uncurry combine ( fidToBDest fid ) ->
           addOp BackupPath
         _ -> do
           let common_l = length $ takeWhile ( uncurry (==) ) $ zip curpath path
           let common_t = length $ takeWhile ( uncurry (==) ) $ zip ( reverse $ drop common_l curpath ) ( reverse $ drop common_l path )
           if common_t >= 2 then
             addOp $ PatchPath ( fromIntegral $ length curpath - common_l ) ( fromIntegral common_t ) $ drop common_l $ reverse $ drop common_t $ reverse path
             else addOp $ RetailPath ( fromIntegral $ length curpath - common_l ) $ drop common_l path


labelStorage :: ( BufferMonad m, Knowledge k ) => String -> StorageId -> F4LMonad k m ()
labelStorage sid label = addOp $ AddStorageLabel sid label


labelEnv :: ( BufferMonad m, Knowledge k ) => String -> EnvironmentId -> F4LMonad k m ()
labelEnv envid label = addOp $ AddEnvironmentLabel envid label


setFid :: ( BufferMonad m, Knowledge k ) => FileId -> F4LMonad k m ()
setFid fid = do
  opEnv <- use st_opEnv
  case fidShorthand fid opEnv of
    Just s -> addOp $ WithKnownFid s
    Nothing -> addOp $ WithFid fid


attrOp :: FilePathAttr a -> Op
attrOp a = case a of
  Require req -> WithPriority req
  Tag tag -> WithTag tag
  Property k v -> WithProperty k v

initDB :: ( FSMonad m, BufferMonad m ) => ErrorT Error ( TopF4LMonad k m ) ()
initDB = lift getDBFile >>= flip appendToFile ""


extraSession :: ( FSMonad m, BufferMonad m ) => TopF4LMonad k m () -> GenF4LMonad k s m ()
extraSession act = do
  pre_s <- use st_session
  state <- get
  end_state <- lift $ execStateT act ( st_session .~ () $ state )
  put $ st_session .~ pre_s $ end_state

inSession :: ( FSMonad m, BufferMonad m, Knowledge k ) 
             => EnvironmentId -> F4LMonad k m a -> GenF4LMonad k s m a
inSession env act = do
  ts <- lift getNow
  sessid <- generateUuid
  inSession' env ts sessid act

inSession' :: ( FSMonad m, BufferMonad m, Knowledge k ) 
             => EnvironmentId -> TimeStamp -> SessionId -> F4LMonad k m a -> GenF4LMonad k s m a
inSession' env ts sessid act = do
  pre_s <- use st_session
  state <- get
  ( res, end_state ) <- lift $ runStateT ( execOp (StartSession undefined undefined env ts 0) >> act >>= \ res -> finalize env ts sessid >> return res ) ( st_session .~ ActiveSession env ts $ state )
  put $ st_session .~ pre_s $ end_state
  return res


finalize :: ( FSMonad m, BufferMonad m, Knowledge k ) 
            => EnvironmentId -> TimeStamp -> SessionId -> F4LMonad k m ()
finalize env ts sessid = do
  l <- bufferLength
  when ( l > 0 ) $ do
    db <- getDBFile
    addOp EndSession
    (buf, bufl) <- freezedBuffer
    onErrorDie $ appendBLsToFile db [ 
        encode ( StartSession currentVersion sessid env ts bufl )
      , BL.pack ( take bufl $ elems buf ) ]
  emptyBuffer


getDBFile :: Monad m => GenF4LMonad k s m FilePath
getDBFile = use ( conf_db . to mdb_file )


loadDB :: FSMonad m => ErrorT Error ( GenF4LMonad k s m ) BL.ByteString
loadDB = do
  f <- lift getDBFile
  processLengthAndBLFromFile' f $ flip const


-- | This re-reads the DB file, so the cache could potentially be assigned a wrong hash
--   in case of concurrent modifications
saveCache :: ( DelFSMonad m, Knowledge k, SerializableKnowledge k ) => ErrorT Error ( TopF4LMonad k m ) ()
saveCache = do
  f <- lift getDBFile
  hs <- liftM ( map $ \ ( _, _, h ) -> h ) $ fromDB $ return . sessionsWithHash
  case reverse hs of
    [] -> return ()
    sha : past -> do
      k <- liftM kCode $ use st_knowledge
      let cfile s = combine ( takeDirectory f ) $ "cache-" ++ k ++ "." ++ hashToS s
      let cf = cfile sha
      p <- lift $ doesFileExist cf
      if p then say Timid ( "Cache file " ++ cf ++ " already exists" )
        else do
        opEnv <- use st_opEnv
        k <- use st_knowledge
        appendBLsToFile cf [ GZ.compress $ runPut $ B.put opEnv >> putK (opEnv ^. openv_fids) k ]
      -- Remove cache files for previous sessions
      lift $ mapM_ ( onErrorUseDefault () . removeFile . cfile ) past
  where
    noK :: Knowledge k1 => k1 -> k1
    noK = const noKnowledge


loadCached :: ( BufferMonad m, DelFSMonad m, ParFSMonad m, SerializableKnowledge k, B.Binary k ) => ErrorT Error ( TopF4LMonad k m ) ()
loadCached = do
  f <- lift getDBFile
  ess <- liftM ( map $ \ (_, a, b) -> (a, b) ) $ fromDB $ return . sessionsWithHash
  k <- liftM kCode $ use st_knowledge
  mc <- lift $ findCache ( takeDirectory f ) k [] $ reverse ess
  case mc of
    Just (cfile, bss) -> do
      let g = do
            openv <- B.get
            k <- getK ( openv ^. openv_fids )
            return ( openv, k )
      outputLineMsg "Loading cache. Please wait..."
      (opEnv, k) <- processLengthAndBLFromFile cfile $ const ( runGet g . GZ.decompress )
      opEnv `deepseq` st_opEnv .= opEnv
      st_knowledge .= k
      let lUncached = sum $ map ( fromIntegral . BL.length ) bss
      when ( lUncached > 100000 ) $ lift $ fork $ onErrorWarn saveCache
      outputLineMsg "Loading ops"
      lift $ loadOps $ BL.concat bss
      cleanLine
    Nothing -> do
      fromDB loadOps
      lift $ fork $ onErrorWarn saveCache
      
  where
    cfile dir k h = combine dir $ "cache-" ++ k ++ "." ++ hashToS h
    findCache :: FSMonad m
                 => FilePath -> String -> [BL.ByteString] -> [( BL.ByteString, SBS.ShortByteString )] -> m ( Maybe ( FilePath, [BL.ByteString] ) )
    findCache _ _ _ [] = return Nothing
    findCache dir kCode pre ( ( bs, h ) : rest ) = do
      let f = cfile dir kCode h
      p <- doesFileExist f
      if p then return $ Just ( f, pre )
        else findCache dir kCode ( bs : pre ) rest


loadOpsDo :: ( FSMonad m, Knowledge k )
             => ( Op -> m () ) -> ( Fact -> m () ) -> BL.ByteString -> TopF4LMonad k m ()
loadOpsDo actOp actFact bs = do
  evalStateT ( ( return $ concatMap snd $ readOps bs ) >>= mapM_ apply >> lift cleanLine ) 0
  where
    apply op = do
      count <- get
      when ( count `mod` 1000 == 0 ) $ lift ( outputLineMsg $ hShow count ++ " ops loaded" )
      put $ count + 1
      facts <- lift $ execOp op
      lift $ lift $ actOp op
      lift $ forM_ facts $ \ fact -> do
        st_knowledge %= applyFact fact
        lift $ actFact fact


sessions :: BL.ByteString -> [(SessionData, BL.ByteString)]
sessions = unfoldr getSession
  where
    getStartSession :: Get ( Maybe SessionData )
    getStartSession = do
      empty <- isEmpty
      if empty then return Nothing else do
        s <- B.get
        case s of
          StartSession a b c d e -> return $ Just $ SessionData a b c d e
          _ -> error "StartSession expected"

    getSession :: BL.ByteString -> Maybe ( ( SessionData, BL.ByteString ), BL.ByteString )
    getSession bs = case runGetOrFail getStartSession bs of
      Right ( ns, offset, Just dat ) ->
        let
          bsize = fromIntegral $ sdat_len dat
        in
         Just ( ( dat, BL.take ( offset + bsize ) bs ), BL.drop bsize ns )
      Right ( _, _, Nothing ) -> Nothing
      Left ( _, _, msg ) -> error msg


sessionsWithHash :: BL.ByteString -> [( SessionData, BL.ByteString, SBS.ShortByteString )]
sessionsWithHash s =
  flip evalState newSha1 $ forM ( sessions s ) $ \ ( sd, bs ) -> do
    modify $ updateSha1 bs
    cs <- get
    return ( sd, bs, toSha1 cs )


fromDB :: FSMonad m => ( BL.ByteString -> GenF4LMonad k s m a ) -> ErrorT Error ( GenF4LMonad k s m ) a
fromDB a = loadDB >>= lift . a


loadOps :: ( FSMonad m, Knowledge k ) => BL.ByteString -> TopF4LMonad k m ()
loadOps = loadOpsDo ( const $ return () ) ( const $ return () )


readOps :: BL.ByteString -> [( SessionData, [Op] )]
readOps = runIdentity . readVOps' ( Identity . toCurrentOp )


readVOps :: BL.ByteString -> [( SessionData, [VOp] )]
readVOps = runIdentity . readVOps' return


readVOps' :: Monad m => ( VOp -> m a ) -> BL.ByteString -> m [( SessionData, [a] )]
readVOps' f s = mapM readSessionOps $ sessions s
  where
    readSessionOps (sd, bs) = do
      res <- case sdat_ver sd of
        V0 -> mapM ( f . VOp0 ) $ readMany bs
        V0_1 -> mapM ( f . VOp0_1 ) $ readMany bs
      return (sd, res)


readMany :: B.Binary b => BL.ByteString -> [b]
readMany =  unfoldr getOne
  where
    mget :: B.Binary b => Get ( Maybe b )
    mget = do
      empty <- isEmpty
      if empty then return Nothing else liftM Just B.get
    getOne bs = case runGetOrFail mget bs of
      Right ( ns, _, Just op ) -> Just ( op, ns )
      Right ( _, _, Nothing ) -> Nothing
      Left ( _, _, msg ) -> error msg


readFullOps :: BL.ByteString -> State OpEnv [( SessionData, [Op] )]
readFullOps = readVOps' $ applyOp' . toCurrentOp
  where
    applyOp' op = do
      applyOp op
      case op of
        WithKnownFid h -> do
          found <- use $ openv_fids . to ( findFid h )
          case found of
            [ fid ] -> return $ WithFid fid
            _ -> error $ "Problem looking for known fid " ++ show h
        _ -> return op 


importSession :: ( FSMonad m, BufferMonad m ) => ( SessionData, [Op] ) -> TopF4LMonad () m ()
importSession ( SessionData ver sessid env ts l, StartSession {} : ops ) =
      inSession' env ts sessid $ mapM_ procOp ops
  where
    procOp :: BufferMonad m => Op -> F4LMonad () m ()
    procOp ( StartSession {} ) = error "StartSession within a session"
    procOp EndSession = return () -- Skip EndSession
    procOp ( WithFid fid ) = do
      use ( st_opEnv . to ( fidShorthand fid ) ) >>= \ sh -> case sh of
        Just s -> addOp $ WithKnownFid s
        Nothing -> addOp $ WithFid fid
    procOp op = addOp op
importSession ( sd,  [] ) = error $ "No op in session " ++ show sd
importSession ( _, ( x : _ ) ) = error $ "Session starts with wrong op " ++ show x


importExportSessions :: ( FSMonad m, BufferMonad m ) 
                        => ( SessionData -> Bool ) -> FilePath -> FilePath -> F4LState () () -> m ()
importExportSessions cond ifile ofile st = do
  whenM ( doesFileExist ofile ) $ error $ "Not proceeding: file " ++ ofile ++ " already exists"
  idat <- onErrorDie $ processLengthAndBLFromFile' ifile $ flip const
  let sops = evalState ( readFullOps idat ) initialOpEnv
  evalStateT ( onErrorDie initDB >> mapM_ importSession ( filter ( cond . fst ) sops ) ) ( conf_db .~ MetaDB ofile $ st )


merge :: ( FSMonad m, BufferMonad m ) => [[( SessionData, [Op] )]] -> TopF4LMonad () m ()
merge = merge' . sortBy ( compare `on` sdat_ts . fst . head ) . filter ( not . null )
  where
    merge' [] = return ()
    merge' ( [] : xs ) = merge' xs
    merge' ( ( x @ ( sd, _ ) : xs ) : xss ) = do
      importSession x
      merge $ xs : map ( dropWhile $ ( == sdat_id sd ) . sdat_id . fst ) xss


mergeDBs :: ( FSMonad m, BufferMonad m ) 
            => [ FilePath ] -> FilePath -> F4LState () () -> m ()
mergeDBs ifiles ofile st = do
  whenM ( doesFileExist ofile ) $ error $ "Not proceeding: file " ++ ofile ++ " already exists"
  idats <- forM ifiles $ \ ifile -> do 
    idat <- onErrorDie $ processLengthAndBLFromFile' ifile $ flip const
    return $ evalState ( readFullOps idat ) initialOpEnv
  evalStateT ( onErrorDie initDB >> merge idats ) ( conf_db .~ MetaDB ofile $ st )
