{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
{-# LANGUAGE FlexibleContexts #-}
module F4L.Interactive where

import F4L
import F4L.Actions
import F4L.BufferMonad
import F4L.FSMonad
import F4L.Knowledge
import F4L.SeenTree as ST
import F4L.Utils hiding( fileSize )

import Control.Lens( at, lens, to, use, (.=), (^.), (.~), Getter, Lens' )
import Control.Monad.Error( runErrorT )
import Control.Monad.State.Strict
import Control.Monad.Trans.Maybe
import qualified Data.Binary as B
import qualified Data.ByteString.Lazy as BL
import Data.Char( isSpace )
import Control.DeepSeq( force )
-- import Data.Either( partitionEithers )
import Data.Function( on )
import Data.List( dropWhileEnd, isPrefixOf, sort, sortBy, foldl' )
import Data.Map.Strict( Map )
import Data.Maybe( listToMaybe )
import Data.Monoid( mappend )
import qualified Data.Map.Strict as SM
import qualified Data.Set as S
import Text.Parsec
import System.Console.Haskeline
-- import System.Console.Haskeline.Completion
-- import System.Console.Haskeline.MonadException
import System.FilePath( joinPath, splitFileName, addTrailingPathSeparator, splitDirectories )


data InteractionContext = Local | InStorage StorageId
                        deriving (Eq, Ord)


-- | Enhanced Knowledge used when performing operations within an interactive session
type InteractionKnowledge =
  ( KnownData
  , ( SeenStorages
    , ( SeenEnvironments
      , ( FactCounter
        , FileSet
        ) ) ) )


icToSid Local = Nothing
icToSid ( InStorage sid ) = Just sid


-- | Structure containing the state built-up during an interactive session
--   containing options etc.
data InteractionStatus = IStatus {
    _is_context :: InteractionContext
  , _is_prev_context :: InteractionContext
  , _is_paths :: Map InteractionContext FilePath
  , _is_quit :: Bool
  , _is_commit :: Bool
  , _is_tags :: [String]
  , _is_props :: [(String, String)]
  }

initialInteractionStatus :: FSMonad m => m InteractionStatus
initialInteractionStatus = do
  cwd <- getCurrentDirectory
  return $ IStatus Local Local (SM.singleton Local cwd) False False [] []

-- is_envid = lens _is_envid $ \ s a -> s { _is_envid = a }

is_context :: Lens' InteractionStatus InteractionContext
is_context = lens _is_context ( \ s a -> s { _is_context = a, _is_prev_context = _is_context s } )

is_prev_context :: Getter InteractionStatus InteractionContext
is_prev_context = to _is_prev_context

is_sid = is_context . to icToSid

is_paths :: Lens' InteractionStatus (Map InteractionContext FilePath)
is_paths = lens _is_paths ( \ s a -> s { _is_paths = a } )

is_path :: Lens' InteractionStatus FilePath
is_path = lens v s
  where
    v x = SM.findWithDefault "/" (_is_context x) (_is_paths x)
    s x a = x { _is_paths = SM.insert (_is_context x) a ( _is_paths x ) }

is_local_path :: Lens' InteractionStatus FilePath
is_local_path = lens v s
  where
    v x = SM.findWithDefault "/" Local (_is_paths x)
    s x a = x { _is_paths = SM.insert Local a ( _is_paths x ) }

{-
is_data :: Lens' InteractionStatus KnownData
is_data = lens _is_data ( \ s a -> s { _is_data = a } )
-}

is_quit :: Lens' InteractionStatus Bool
is_quit = lens _is_quit ( \ s a -> s { _is_quit = a } )

is_commit :: Lens' InteractionStatus Bool
is_commit = lens _is_commit ( \ s a -> s { _is_commit = a } )

byContext :: Monad m => Interaction m a -> ( StorageId -> Interaction m a ) -> Interaction m a
byContext whenLocal whenStorage =
  use is_sid >>= maybe whenLocal whenStorage


joinDirCanonical :: FSMonad m => String -> String -> m String
joinDirCanonical parent child = canonicalizePath $ joinPath [ parent, child ]

joinDir :: String -> String -> String
joinDir parent child = joinPath $ j [] $ filter ( /= "." ) $ splitDirectories parent ++ splitDirectories child
    where
      j pre ( "/" : ".." : rest ) = j pre ( "/" : rest )
      j pre ( _x : ".." : rest ) = j ( pre ++ rest ) []
      j pre [] = pre
      j pre ( x : rest ) = j ( pre ++ [ x ] ) rest


type Interaction m = StateT InteractionStatus m


sidContents :: ( Monad m, KnownDataClass k ) => StorageId -> [FilePath] -> F4LMonad k m ( Maybe SeenTree )
sidContents sid path = do
  contents <- use $ st_knowledge . kd_contents . to ( SM.findWithDefault ST.empty sid )
  return $ findNode path contents


listPaths :: FSMonad m
             => String -> Interaction ( F4LMonad InteractionKnowledge m ) ([(Bool, Completion)], [(Bool, Completion)])
listPaths pre = do
  cwd <- use is_path
  let (dir, file) = splitFileName pre
  local <- do
    let rdir = joinPath [cwd, dir]
    dc <- onErrorUseDefault [] ( getDirectoryContents rdir ) >>= return . filter ( file `isPrefixOf` )
    forM dc $ \ f -> do
      let p = joinPath $ [ d | d <- [dir], d /= dummyDir ] ++ [ f ]
      isDir <- doesDirectoryExist $ joinPath [cwd, p]
      return $ if isDir then dirComplete p else fileComplete p
  current <- byContext ( return local ) $ \ sid -> do
      let rdir = dropWhile ( == "/" ) $ splitDirectories $ joinDir cwd dir
      node <- lift $ sidContents sid rdir
      case node of
        Just ( Dir st ) ->
          return $ do -- list monad
            (subPath, node) <- map ( mapFst shortToPath ) $ SM.toList st
            let p = joinPath $ [ d | d <- [dir], d /= dummyDir ] ++ [ subPath ]
            if file `isPrefixOf` subPath then (
               case node of 
                 File _ -> return $ fileComplete p
                 Dir _ -> return $ dirComplete p
               ) else []
        _ -> return []
  return (local, current)
  
  where
    dummyDir = fst $ splitFileName "abc" -- TODO: de-uglify

    fileComplete f = (False, Completion f f True)

    dirComplete d = let td = addTrailingPathSeparator d in (True, Completion td td False)


-- | Type of completion to perform when parsing input as an incomplete command
data Complete = CompletePath
              | CompleteLocalPath
              | CompleteDir
              | CompleteLocalDir
              | CompleteSid
              | CompleteEnvid
              | CompleteFixed [(String, Maybe String)] -- ^ ( completion, help )
              deriving (Eq, Show, Ord)


iComplete :: (MonadException m, DelFSMonad m, BufferMonad m) 
             => CompletionFunc ( Interaction ( F4LMonad InteractionKnowledge m ) )
iComplete (rline, _) = do
  parsed <- runParserT (toCompletionParser parseCmd) () "" $ reverse rline
  -- outputMsg $ show parsed
  let posParses = either (const []) id parsed
  let pos = foldr max 0 $ map fst posParses
  let part = drop pos $ reverse rline
  -- Only take maximal parses
  let parses = map snd $ filter ( (== pos) . fst ) posParses
  storages <- lift $ use $ st_knowledge . seen_storages
  environments <- lift $ use $ st_knowledge . seen_environments
  ws @ (_, _) <- listPaths part
  return ( reverse $ take pos $ reverse rline
         , nubSort $ join $ map ( getComp storages environments ws part ) $ sNub $ sort parses )
  where
    nubSort :: [Completion] -> [Completion]
    nubSort = sNubBy sameReplacement . sortBy cmpCompletion

    sameReplacement = (==) `on` replacement

    cmpCompletion = ( compare `on` replacement ) `mappend` ( compare `on` isFinished )

    getFiniteComp part ss =
      [ Completion a ( maybe b ( ( b ++ "\t" ) ++ ) c ) True
      | ( a, b, c ) <- ss
      , part `isPrefixOf` a ]

    getComp _ _ _ part (CompleteFixed ss) = getFiniteComp part $ map ( \ (a, b) -> (a, a, b) ) ss
    getComp _ _ (_, ws) _ CompleteDir = map snd $ filter fst ws
    getComp _ _ (ws, _) _ CompleteLocalDir = map snd $ filter fst ws
    getComp _ _ (_, ws) _ CompletePath = map snd ws
    getComp _ _ (ws, _) _ CompleteLocalPath = map snd ws
    getComp storages _ _ part CompleteSid =
      getFiniteComp part [ (k, abbrSid k, listToMaybe v) | (k, v) <- SM.toList storages ]
    getComp _ environments _ part CompleteEnvid = 
      getFiniteComp part [ (k, k, listToMaybe v) | (k, v) <- SM.toList environments ]

  
(<+>) :: Stream s m t => ParsecT s u m [a] -> ParsecT s u m [a] -> ParsecT s u m [a]
a <+> b = do
  ares <- try ( lookAhead a ) <|> return []
  bres <- try ( lookAhead b ) <|> return []
  return $ ares ++ bres


getPos :: (Monad m, Stream String m Char) => ParsecT String u m Int
getPos = getParserState >>= return . (\ x -> x - 1) . sourceColumn . statePos


-- | Combination of a partial and a full parser
--   This is used to define in one go how to interpret a command
--   and what tab-completion to perform
data Combo u m a = Combo {
    combo_partial :: ParsecT String u m [(Int, Complete)]
  , combo_full :: ParsecT String u m a
  }

terminated :: Monad m => Combo u m a -> Combo u m a
terminated ( Combo a b ) = Combo a ( b <<- many ( char ' ' ) <<- eof )


instance (Monad m) => Monad ( Combo u m ) where
  Combo a1 a2 >>= f =
    Combo
      ( try a1 <|> ( a2 >>= combo_partial . f ) )
      ( a2 >>= combo_full . f)
  return a = Combo (return []) (return a)


instance MonadTrans ( Combo u ) where
  lift a = Combo ( return [] ) ( lift a )


someString a = someStringHelp' a Nothing

someStringHelp :: Monad m => String -> String -> Combo u m String
someStringHelp a b = someStringHelp' a ( Just b )

someStringHelp' :: Monad m => String -> Maybe String -> Combo u m String
someStringHelp' s help = Combo p f
  where
    p = do
      pos <- getPos
      many $ noneOf " "
      eof
      return [(pos, CompleteFixed [(s, help)])]
    f = string s


somePath_ :: Monad m => Complete -> Combo u m String
somePath_ c = Combo p f
  where
    p = do
      pos <- getPos
      many $ noneOf " "
      eof
      return [(pos, c)]
    f = many1 ( noneOf " " )


somePath :: Monad m => Combo u m String
somePath = somePath_ CompletePath

someLocalPath :: Monad m => Combo u m String
someLocalPath = somePath_ CompleteLocalPath

someDir :: Monad m => Combo u m String
someDir = somePath_ CompleteDir

someLocalDir :: Monad m => Combo u m String
someLocalDir = somePath_ CompleteLocalDir

someSid :: Monad m => Combo u m String
someSid = Combo p f
  where
    p = do
      pos <- getPos
      many $ noneOf " "
      eof
      return [(pos, CompleteSid)]
    f = many1 $ noneOf " "

someEnvid :: Monad m => Combo u m String
someEnvid = Combo p f
  where
    p = do
      pos <- getPos
      many $ noneOf " "
      eof
      return [(pos, CompleteEnvid)]
    f = many1 $ noneOf " "


someSize :: Monad m => Combo u m Int
someSize = Combo ( fail "No parse" ) $ do
  i <- many1 ( oneOf ['0' .. '9'] )
  s <- liftM ( : [] ) ( try ( oneOf "kKmMgG" ) ) <|> return ""
  return $ readN $ i ++ s


someVerbosity :: Monad m => Combo u m VerbosityLevel
someVerbosity = foldl1 orElse [ someString ( show i ) >> return ( toEnum i ) | i <- [0 .. 3] ]


somePriority :: Monad m => Combo u m Priority
somePriority = 
  foldl1 orElse $ [
      someStringHelp "none" "don't request backups" >> return Skip
    , someStringHelp "auto" "standard" >> return DefaultPriority
    , someStringHelp "top" "highest priority" >> return AsManyAsPossible
    ] ++ [ someString ( show i ) >> return ( Priority i ) | i <- [1..10] ]


someBool :: Monad m => Combo u m Bool
someBool = foldl1 orElse [
    someString "true" >> return True
  , someString "false" >> return False
  ]


someLabel :: Monad m => Combo u m String
someLabel = Combo p f
  where
    p = return []
    f = liftM ( dropWhile isSpace . dropWhileEnd isSpace ) $ liftM2 (:) ( noneOf " " ) ( many1 anyChar )


condM' :: Monad m => m ( Maybe c ) -> ( c -> Combo u m a ) -> Combo u m a
condM' cond f = Combo ( cond' ( combo_partial .f ) ) ( cond' ( combo_full . f ) )
  where
    cond' g = do
      c <- lift cond
      case c of
        Nothing -> fail ""
        Just v -> g v


condM a f = condM' b ( const f )
  where
    b = do
      v <- a
      return $ if v then Just () else Nothing


Combo a1 a2 `orElse` Combo b1 b2 = Combo (a1 <+> b1) (try a2 <|> b2)


Combo a1 a2 `andDo` f2 = Combo a1 (liftM f2 a2)


a <<- b = do
  res <- a
  b
  return res


sp' ( Combo a1 a2 ) f =
  Combo
    ( try a1 <|> ( ( a2 <<- many1 (char ' ') ) >>= combo_partial . f ) )
    ( ( a2 <<- many1 (char ' ') ) >>= combo_full . f )

sp a = sp' a . const


several :: Monad m => Combo u m String -> Combo u m [ String ]
several a = ( a `sp'` \ x -> liftM ( x : ) $ several a ) `orElse` liftM ( : [] ) a


toFullParser :: Stream String m Char => Combo u m a -> ParsecT String u m a
toFullParser a = combo_full a <<- many ( char ' ' ) <<- eof


toCompletionParser :: Stream String m Char => Combo u m (m ()) -> ParsecT String u m [(Int, Complete)]
toCompletionParser = combo_partial


ls :: ( FSMonad m, KnownDataClass k ) => String -> Interaction (F4LMonad k m) ()
ls rel = do
  dir <- byContext ( 
    do
      path <- use is_path >>= lift . flip joinDirCanonical rel
      lift $ onErrorWarnWithDefault [] $ getDirectoryContents path 
    ) ( \ sid -> do
           path <- use is_path >>= return . dropWhile (== "/") . splitDirectories . flip joinDir rel
           contents <- lift $ sidContents sid path
           case contents of
             Just ( Dir st ) -> return $ map shortToPath $ SM.keys st
             Just ( File _ ) -> return [joinPath path]
             _ -> return []
      )
  -- TODO: show whether file or dir
  mapM_ outputMsg $ sort dir


-- | change relative path to path relative to the storage's /
storagePath :: Monad m => FilePath -> Interaction ( F4LMonad InteractionKnowledge m ) FilePath
storagePath p = use is_path >>= return . dropWhile (== '/') . flip joinDir p

iRestore :: ( BufferMonad m, DelFSMonad m )
            => Bool -> StorageId -> FilePath -> FilePath -> Interaction ( F4LMonad InteractionKnowledge m ) ()
iRestore record sid p lp = do
  path <- storagePath p
  lpath <- use is_local_path >>= lift . flip joinDirCanonical lp
  eres <- lift $ outputMsg ( show (sid, path, lpath) ) >> runErrorT ( restore record sid path lpath )
  case eres of
    Left err ->
      outputMsg $ "Error: " ++ err
    Right res -> do
      let acc (_, _, Nothing) (i, j, k) = (i+1, j, k)
          acc (_, _, Just False) (i, j, k) = (i, j+1, k)
          acc (_, _, Just True) (i, j, k) = (i, j, k+1)
      let (skipped, notFound, restored) = foldr acc (0, 0, 0) res
      outputMsg $ show restored ++ " files restored"
      outputMsg $ show skipped ++ " files already present"
      outputMsg $ show notFound ++ " files not found"


cd :: FSMonad m => String -> Interaction m ()
cd newdir = byContext ( 
  do
    newPath <- use is_path >>= lift . flip joinDirCanonical newdir
    whenM ( lift $ doesDirectoryExist newPath ) $ is_path .= newPath
  ) $ const ( 
  do
    newPath <- use is_path >>= return . flip joinDir newdir
    is_path .= newPath
  )


showSettings :: ( FSMonad m, KnownDataClass k ) => Interaction ( F4LMonad k m ) ()
showSettings = do
  withRescan <- lift $ use st_withRescan
  lift $ outputMsg $ "Rescan: " ++ show withRescan
  verbosity <- lift $ use st_verbosity
  lift $ outputMsg $ "Verbosity level: " ++ show ( verbosity ^. verb_level )
  lift $ outputMsg $ "Tracing: " ++ show ( verbosity ^. verb_tracing )


showSessions :: FSMonad m => Interaction ( F4LMonad k m ) ()
showSessions = lift $ do
  outputMsg "Warning: uncommited data will not appear"
  onErrorWarn $ fromDB ( return . sessions ) >>= mapM_ ( lift . outputMsg . formatSession . fst )


abbrSid :: StorageId -> String
abbrSid sid = take 6 sid ++ "..."

showStorages :: ( SeenStoragesClass k, FSMonad m ) => Interaction ( F4LMonad k m ) ()
showStorages = lift $ do
  ss <- use $ st_knowledge . seen_storages
  mapM_ s' $ sort $ SM.toList ss
  where
    s' ( sid, x : _ ) = outputMsg $ abbrSid sid ++ " " ++ x
    s' ( sid, _ ) = outputMsg $ abbrSid sid


changeToStorage :: ( SeenStoragesClass k, FSMonad m ) => StorageId -> Interaction ( F4LMonad k m ) ()
changeToStorage newsid = do
  ss <- lift $ use $ st_knowledge . seen_storages
  case filter ( newsid `isPrefixOf` ) $ SM.keys ss of
    [ sid ] -> is_context .= InStorage sid
    [] -> outputMsg $ "No storage id starting with " ++ newsid
    x -> outputMsg $ show ( length x ) ++ " storages starting with " ++ newsid ++ ". Use a more specific id"


infoPath :: ( FSMonad m, BufferMonad m ) 
            => StorageId -> FilePath -> Interaction ( F4LMonad InteractionKnowledge m ) ()
infoPath sid p = do
  path <- storagePath p
  runMaybeT $ do
    tree <- MaybeT $ lift $ use $ st_knowledge . kd_contents . at sid
    fid <- MaybeT $ return $ findFileNode path tree
    lift $ infoFid fid
  return ()


infoLocalPath :: ( FSMonad m, BufferMonad m ) 
            => FilePath -> Interaction ( F4LMonad InteractionKnowledge m ) ()
infoLocalPath p = do
  path <- use is_local_path >>= lift . flip joinDirCanonical p
  mfid <- onErrorWarnWithDefault Nothing $ liftM Just $ fileId path
  case mfid of
    Just fid -> infoFid fid
    Nothing -> return ()


infoFid :: ( FSMonad m, BufferMonad m )
           => FileId -> Interaction ( F4LMonad InteractionKnowledge m ) ()
infoFid fid = do
  outputMsg $ show fid
  candidates <- lift $ findLocalFileCandidates fid
  localPaths <- filterM ( liftM ( == fid_length fid ) . onErrorUseDefault (-1) . fileSize ) candidates
  unless ( null localPaths ) $ do
    outputMsg "Locally at:"
    mapM_ outputMsg localPaths
  mfdat <- lift $ use $ st_knowledge . kd_files . at' fid
  case mfdat of
    Just fdat -> do
      outputMsg $ "Priority: " ++ show ( fdat ^. fd_priority )
      forM_ ( sort $ map ( mapSnd shortToPath ) $ S.toList $ fdat ^. fd_seen ) $ \ ( sid, path ) -> do
        outputMsg $ "Seen in " ++ sid ++ "\t" ++ path
    Nothing -> return ()


infoStorage :: FSMonad m => StorageId -> Interaction ( F4LMonad InteractionKnowledge m ) ()
infoStorage sid = lift $ do
  outputMsg $ "Info for storage " ++ sid
  ss <- use $ st_knowledge . kd_storages . to ( SM.findWithDefault [] sid )
  forM_ ss $ \ ( envid, path ) -> outputMsg ( "Seen on env " ++ envid ++ " in " ++ path )
  envid <- use st_envid
  let seenLocally = map snd $ filter ( (== envid) . fst ) ss
  findStorage sid seenLocally >>= mapM_ ( outputMsg . ( "Currently on " ++ ) )
  st <- use $ st_knowledge . kd_contents . to ( SM.findWithDefault ST.empty sid )
  let (n, l) = foldl' ( \ (n, l) fid -> ( force $ n + 1, force $ l + fid_length fid ) ) (0, 0) $ fids st
  outputMsg $ hShow n ++ " files"
  outputMsg $ hShow l ++ " bytes"


infoEnvironment :: FSMonad m => EnvironmentId -> Interaction ( F4LMonad InteractionKnowledge m ) ()
infoEnvironment envid = return () -- To be implemented


showStats :: ( FSMonad m, BufferMonad m ) => F4LMonad InteractionKnowledge m ()
showStats = do
  s <- stats
  sequence_ $ do
    ( label, dat ) <- s
    return $ lift $ outputMsg $ label ++ "\t" ++ dat


debug :: FSMonad m => Interaction ( F4LMonad InteractionKnowledge m ) ()
debug = do
  contents <- lift $ use $ st_knowledge . kd_contents
  outputMsg $ show $ length $ show $ concatMap allFiles $ SM.elems contents
  files <- lift $ use $ st_knowledge . kd_files
  outputMsg $ show $ length $ show files
  outputMsg $ show $ BL.length $ B.encode contents
  outputMsg $ show $ sum [ 1 .. 1000000 ]


parseCmd :: ( Stream String n Char, DelFSMonad m, BufferMonad m )
            => Combo u (Interaction n) ( Interaction ( F4LMonad InteractionKnowledge m ) () )
parseCmd = foldr1 orElse $ map terminated [
    someString "quit" `andDo` const ( is_quit .= True )
  , someString "commit" `andDo` const ( is_commit .= True )
  , ( someString "ls" `sp` somePath ) `andDo` ls
  , someString "ls" `andDo` const ( ls "" )
  , ( someString "cd" `sp` someDir ) `andDo` cd
  , ( csHelp `sp` someStringHelp "-" "previous storage" ) `andDo` const ( use is_prev_context >>= (.=) is_context )
  , ( csHelp `sp` someSid ) `andDo` changeToStorage
  , ( csHelp ) `andDo` const ( is_context .= Local )
  , someString "scan" `sp` someString "--priority" `sp` liftM Just somePriority `sp'` scan'
  , someString "scan" `sp` scan' Nothing
  , someString "show" `sp` someString "coverage" `andDo` const ( lift $ coverage >>= mapM_ ( outputMsg . show ) )
  , someString "show" `sp` someString "stats" `andDo` const ( lift showStats )
  , someString "show" `sp` someString "settings" `andDo` const showSettings
  , someString "show" `sp` someString "sessions" `andDo` const showSessions
  , someString "show" `sp` someString "storages" `andDo` const showStorages
  , inStorage $ \ sid -> someString "info" `sp` someString "--file" `sp` somePath `andDo` infoPath sid
  , inLocal $ someString "info" `sp` someString "--file" `sp` somePath `andDo` infoLocalPath
  , someString "info" `sp` someString "--storage" `sp` someSid `andDo` infoStorage
  , inStorage $ \ sid -> someString "info" `sp` someString "--storage" `andDo` const ( infoStorage sid )
  , someString "info" `sp` someString "--env" `sp` someEnvid `andDo` infoEnvironment
  , inStorage $ \ sid -> restoreHelp `sp` someString "--track" `sp` somePath `sp'` \ p -> someLocalPath >>= \ lp -> return ( iRestore True sid p lp )
  , inStorage $ \ sid -> restoreHelp `sp` someString "--notrack" `sp` somePath `sp'` \ p -> someLocalPath >>= \ lp -> return ( iRestore False sid p lp )
  , backupHelp
    `sp` someString "-s"
    `sp` someSize
    `sp'` \ s -> someLocalDir `andDo` ( lift . flip autoBackup s )
  , inStorage $ \ sid -> backupHelp
    `sp` somePath
    `sp'` \ s -> someLocalDir `andDo` backupP sid s
  , someString "set" `sp` someString "rescan" `sp` someBool `andDo` ( lift . ( st_withRescan .= ) )
  , someString "set" `sp` someString "verbosity" `sp` someVerbosity `andDo` ( lift . ( st_verbosity . verb_level .= ) )
  , someString "debug" `andDo` const debug
  , someString "label" `sp` someString "--environment" `sp` someEnvid `sp'` \ env -> someLabel >>= \ label -> return ( lift $ labelEnv env label )
  , someString "label" `sp` someString "--storage" `sp` someSid `sp'` \ sid -> someLabel >>= \ label -> return ( lift $ labelStorage sid label )
  ]
  where
    isLocal Local = True
    isLocal _ = False
    inLocal :: Monad m => Combo u ( Interaction m ) a -> Combo u ( Interaction m ) a
    inLocal = condM $ use $ is_context . to isLocal
    inStorage = condM' $ use $ is_context . to icToSid
    scan' mpri = several someLocalPath `andDo` \ paths -> do
        cwd <- use is_local_path
        toAdd <- mapM ( joinDirCanonical cwd ) paths
        lift $ addFiles mpri toAdd
    csHelp = someStringHelp "cs" "change current storage (no arg for local, '-' for previous)"
    backupHelp = someStringHelp "backup" "save some files to backup storage"
    restoreHelp = someStringHelp "restore" "restore to local directory" 
    backupP sid source dest = do
      rsource <- use is_path >>= \ d -> return $ dropWhile ( == '/' ) ( joinDir d source )
      lift $ backupPath sid rsource dest

runShell :: (MonadException m, DelFSMonad m, BufferMonad m)
            => EnvironmentId -> TopF4LMonad InteractionKnowledge m ()
runShell envid = do
  init <- initialInteractionStatus
  externalLoop Nothing init
    where
      externalLoop mhist st = do
        ( newHist, newSt ) <- inSession envid $ flip runStateT st $ runInputT ( setComplete iComplete defaultSettings ) $ do
          case mhist of 
            Just hist -> putHistory hist 
            _ -> return ()
          internalLoop 
          getHistory
        case ( newSt ^. is_commit, newSt ^. is_quit ) of
          ( True, False ) -> externalLoop ( Just newHist ) ( is_quit .~ False $ newSt )
          ( _, True ) -> return ()
          ( False, False ) -> error "Abnormal exit"
      internalLoop = do
        msid <- lift $ use is_sid
        path <- lift $ use is_path
        
        let prompt = maybe "" (\ x -> "[" ++ abbrSid x ++ "] ") msid ++ path ++ " > "
        minput <- getInputLine prompt

        case minput of
          Nothing -> 
            lift $ is_quit .= True
          Just line -> do
            parse <- lift $ runParserT (combo_full parseCmd) () "" line
            case parse of
              Left err ->
                outputStrLn $ "Unable to parse command " ++ show err
              Right act ->
                lift act
        toQuit <- lift $ use is_quit
        toCommit <- lift $ use is_commit
        unless ( toQuit || toCommit ) internalLoop
